@@ -11,12 +11,12 @@
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
-  
+
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
-  
+
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */
@@ -52,13 +52,17 @@
 #endif
 
 static pinentry_t pinentry;
+static int grab_failed;
 static int passphrase_ok;
-static int confirm_yes;
+typedef enum { CONFIRM_CANCEL, CONFIRM_OK, CONFIRM_NOTOK } confirm_value_t;
+static confirm_value_t confirm_value;
 
 static GtkWidget *entry;
 static GtkWidget *qualitybar;
+#ifdef ENABLE_ENHANCED
 static GtkWidget *insure;
 static GtkWidget *time_out;
+#endif
 static GtkTooltips *tooltips;
 
 /* Gnome hig small and large space in pixels.  */
@@ -66,10 +70,10 @@ static GtkTooltips *tooltips;
 #define HIG_LARGE     12
 
 /* The text shown in the quality bar when no text is shown.  This is not
- * the empty string, becase because with an empty string the height of
+ * the empty string, because with an empty string the height of
  * the quality bar is less than with a non-empty string.  This results
  * in ugly layout changes when the text changes from non-empty to empty
- * and vice versa */
+ * and vice versa.  */
 #define QUALITYBAR_EMPTY_TEXT " "
 
 
@@ -94,28 +98,70 @@ constrain_size (GtkWidget *win, GtkRequisition *req, gpointer data)
 }
 
 
+/* Realize the window as transient if we grab the keyboard.  This
+   makes the window a modal dialog to the root window, which helps the
+   window manager.  See the following quote from:
+   http://standards.freedesktop.org/wm-spec/wm-spec-1.4.html#id2512420
+
+   Implementing enhanced support for application transient windows
+
+   If the WM_TRANSIENT_FOR property is set to None or Root window, the
+   window should be treated as a transient for all other windows in
+   the same group. It has been noted that this is a slight ICCCM
+   violation, but as this behavior is pretty standard for many
+   toolkits and window managers, and is extremely unlikely to break
+   anything, it seems reasonable to document it as standard.  */
+
+static void
+make_transient (GtkWidget *win, GdkEvent *event, gpointer data)
+{
+  GdkScreen *screen;
+  GdkWindow *root;
+
+  if (! pinentry->grab)
+    return;
+
+  /* Make window transient for the root window.  */
+  screen = gdk_screen_get_default ();
+  root = gdk_screen_get_root_window (screen);
+  gdk_window_set_transient_for (win->window, root);
+}
+
+
 /* Grab the keyboard for maximum security */
 static void
 grab_keyboard (GtkWidget *win, GdkEvent *event, gpointer data)
 {
-  if (!pinentry->grab)
+  if (! pinentry->grab)
     return;
 
   if (gdk_keyboard_grab (win->window, FALSE, gdk_event_get_time (event)))
-    g_error ("could not grab keyboard");
+    {
+      g_critical ("could not grab keyboard");
+      grab_failed = 1;
+      gtk_main_quit ();
+    }
 }
 
+
 /* Remove grab.  */
 static void
 ungrab_keyboard (GtkWidget *win, GdkEvent *event, gpointer data)
 {
   gdk_keyboard_ungrab (gdk_event_get_time (event));
+  /* Unmake window transient for the root window.  */
+  /* gdk_window_set_transient_for cannot be used with parent = NULL to
+     unset transient hint (unlike gtk_ version which can).  Replacement
+     code is taken from gtk_window_transient_parent_unrealized.  */
+  gdk_property_delete (win->window,
+                       gdk_atom_intern_static_string ("WM_TRANSIENT_FOR"));
 }
 
 
 static int
 delete_event (GtkWidget *widget, GdkEvent *event, gpointer data)
 {
+  pinentry->close_button = 1;
   gtk_main_quit ();
   return TRUE;
 }
@@ -129,12 +175,15 @@ button_clicked (GtkWidget *widget, gpointer data)
       const char *s;
 
       /* Okay button or enter used in text field.  */
-
+#ifdef ENABLE_ENHANCED
+      /* FIXME: This is not compatible with assuan.  We can't just
+	 print stuff on stdout.  */
       if (pinentry->enhanced)
 	printf ("Options: %s\nTimeout: %d\n\n",
 		gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (insure))
 		? "insure" : "",
 		gtk_spin_button_get_value_as_int (GTK_SPIN_BUTTON (time_out)));
+#endif
 
       s = gtk_secure_entry_get_text (GTK_SECURE_ENTRY (entry));
       if (!s)
@@ -158,10 +207,7 @@ enter_callback (GtkWidget *widget, GtkWidget *anentry)
 static void
 confirm_button_clicked (GtkWidget *widget, gpointer data)
 {
-  if (data)
-    /* Okay button.  */
-    confirm_yes = 1;
-
+  confirm_value = (int) data;
   gtk_main_quit ();
 }
 
@@ -177,6 +223,8 @@ pinentry_utf8_validate (gchar *text)
   if (g_utf8_validate (text, -1, NULL))
     return g_strdup (text);
 
+  /* Failure: Assume that it was encoded in the current locale and
+     convert it to utf-8.  */
   result = g_locale_to_utf8 (text, -1, NULL, NULL, NULL);
   if (!result)
     {
@@ -225,7 +273,7 @@ changed_text_handler (GtkWidget *widget)
       snprintf (textbuf, sizeof textbuf, "%d%%", percent);
       color.green = 0xffff;
     }
-  gtk_progress_bar_set_fraction (GTK_PROGRESS_BAR (qualitybar), 
+  gtk_progress_bar_set_fraction (GTK_PROGRESS_BAR (qualitybar),
                                  (double)percent/100.0);
   gtk_progress_bar_set_text (GTK_PROGRESS_BAR (qualitybar), textbuf);
   gtk_widget_modify_bg (qualitybar, GTK_STATE_PRELIGHT, &color);
@@ -260,15 +308,26 @@ create_window (int confirm_mode)
 		    G_CALLBACK (constrain_size), NULL);
   if (!confirm_mode)
     {
+      if (pinentry->grab)
+	g_signal_connect (G_OBJECT (win),
+			  "realize", G_CALLBACK (make_transient), NULL);
+
+      /* We need to grab the keyboard when its visible! not when its
+         mapped (there is a difference)  */
+      g_object_set (G_OBJECT(win), "events",
+                    GDK_VISIBILITY_NOTIFY_MASK | GDK_STRUCTURE_MASK, NULL);
+
       g_signal_connect (G_OBJECT (win),
-			pinentry->grab ? "map-event" : "focus-in-event",
+			pinentry->grab
+                        ? "visibility-notify-event"
+                        : "focus-in-event",
 			G_CALLBACK (grab_keyboard), NULL);
       g_signal_connect (G_OBJECT (win),
 			pinentry->grab ? "unmap-event" : "focus-out-event",
 			G_CALLBACK (ungrab_keyboard), NULL);
     }
   gtk_window_add_accel_group (GTK_WINDOW (win), acc);
-  
+
   wvbox = gtk_vbox_new (FALSE, HIG_LARGE * 2);
   gtk_container_add (GTK_CONTAINER (win), wvbox);
   gtk_container_set_border_width (GTK_CONTAINER (wvbox), HIG_LARGE);
@@ -284,6 +343,11 @@ create_window (int confirm_mode)
   box = gtk_vbox_new (FALSE, HIG_SMALL);
   gtk_box_pack_start (GTK_BOX (chbox), box, TRUE, TRUE, 0);
 
+  if (pinentry->title)
+    {
+      msg = pinentry_utf8_validate (pinentry->title);
+      gtk_window_set_title (GTK_WINDOW(win), msg);
+    }
   if (pinentry->description)
     {
       msg = pinentry_utf8_validate (pinentry->description);
@@ -317,7 +381,7 @@ create_window (int confirm_mode)
       if (pinentry->prompt)
 	{
 	  msg = pinentry_utf8_validate (pinentry->prompt);
-	  w = gtk_label_new (msg);
+	  w = gtk_label_new_with_mnemonic (msg);
 	  g_free (msg);
 	  gtk_misc_set_alignment (GTK_MISC (w), 1.0, 0.5);
 	  gtk_table_attach (GTK_TABLE (table), w, 0, 1, 0, 1,
@@ -356,6 +420,7 @@ create_window (int confirm_mode)
 	  		    GTK_EXPAND|GTK_FILL, GTK_EXPAND|GTK_FILL, 0, 0);
 	}
 
+#ifdef ENABLE_ENHANCED
       if (pinentry->enhanced)
 	{
 	  GtkWidget *sbox = gtk_hbox_new (FALSE, HIG_SMALL);
@@ -370,7 +435,7 @@ create_window (int confirm_mode)
 	     2, 0);
 	  gtk_box_pack_start (GTK_BOX (sbox), time_out, FALSE, FALSE, 0);
 	  gtk_widget_show (time_out);
-	  
+
 	  w = gtk_label_new ("seconds");
 	  gtk_box_pack_start (GTK_BOX (sbox), w, FALSE, FALSE, 0);
 	  gtk_widget_show (w);
@@ -381,6 +446,7 @@ create_window (int confirm_mode)
 	  gtk_box_pack_start (GTK_BOX (box), insure, FALSE, FALSE, 0);
 	  gtk_widget_show (insure);
 	}
+#endif
     }
 
   bbox = gtk_hbutton_box_new ();
@@ -393,24 +459,62 @@ create_window (int confirm_mode)
       if (pinentry->cancel)
         {
           msg = pinentry_utf8_validate (pinentry->cancel);
-          w = gtk_button_new_with_label (msg);
+          w = gtk_button_new_with_mnemonic (msg);
           g_free (msg);
         }
+      else if (pinentry->default_cancel)
+        {
+          GtkWidget *image;
+
+          msg = pinentry_utf8_validate (pinentry->default_cancel);
+          w = gtk_button_new_with_mnemonic (msg);
+          g_free (msg);
+          image = gtk_image_new_from_stock (GTK_STOCK_CANCEL,
+                                            GTK_ICON_SIZE_BUTTON);
+          if (image)
+            gtk_button_set_image (GTK_BUTTON (w), image);
+        }
       else
-        w = gtk_button_new_from_stock (GTK_STOCK_CANCEL);
+          w = gtk_button_new_from_stock (GTK_STOCK_CANCEL);
       gtk_container_add (GTK_CONTAINER (bbox), w);
       g_signal_connect (G_OBJECT (w), "clicked",
                         G_CALLBACK (confirm_mode ? confirm_button_clicked
-                                    : button_clicked), NULL);
+                                    : button_clicked),
+			(gpointer) CONFIRM_CANCEL);
       GTK_WIDGET_SET_FLAGS (w, GTK_CAN_DEFAULT);
     }
-  
+
+  if (confirm_mode && !pinentry->one_button && pinentry->notok)
+    {
+      msg = pinentry_utf8_validate (pinentry->notok);
+      w = gtk_button_new_with_mnemonic (msg);
+      g_free (msg);
+
+      gtk_container_add (GTK_CONTAINER (bbox), w);
+      g_signal_connect (G_OBJECT (w), "clicked",
+                        G_CALLBACK (confirm_button_clicked),
+			(gpointer) CONFIRM_NOTOK);
+      GTK_WIDGET_SET_FLAGS (w, GTK_CAN_DEFAULT);
+    }
+
   if (pinentry->ok)
     {
       msg = pinentry_utf8_validate (pinentry->ok);
-      w = gtk_button_new_with_label (msg);
+      w = gtk_button_new_with_mnemonic (msg);
       g_free (msg);
     }
+  else if (pinentry->default_ok)
+    {
+      GtkWidget *image;
+
+      msg = pinentry_utf8_validate (pinentry->default_ok);
+      w = gtk_button_new_with_mnemonic (msg);
+      g_free (msg);
+      image = gtk_image_new_from_stock (GTK_STOCK_OK,
+                                        GTK_ICON_SIZE_BUTTON);
+      if (image)
+        gtk_button_set_image (GTK_BUTTON (w), image);
+    }
   else
     w = gtk_button_new_from_stock (GTK_STOCK_OK);
   gtk_container_add (GTK_CONTAINER(bbox), w);
@@ -427,14 +531,16 @@ create_window (int confirm_mode)
   else
     {
       g_signal_connect (G_OBJECT (w), "clicked",
-			G_CALLBACK(confirm_button_clicked), "ok");
+			G_CALLBACK(confirm_button_clicked),
+			(gpointer) CONFIRM_OK);
       GTK_WIDGET_SET_FLAGS (w, GTK_CAN_DEFAULT);
     }
 
   gtk_window_set_position (GTK_WINDOW (win), GTK_WIN_POS_CENTER);
-  
-  gtk_widget_show_all(win);
-  
+  gtk_window_set_keep_above (GTK_WINDOW (win), TRUE);
+  gtk_widget_show_all (win);
+  gtk_window_present (GTK_WINDOW (win));  /* Make sure it has the focus.  */
+
   return win;
 }
 
@@ -446,7 +552,7 @@ gtk_cmd_handler (pinentry_t pe)
   int want_pass = !!pe->pin;
 
   pinentry = pe;
-  confirm_yes = 0;
+  confirm_value = CONFIRM_CANCEL;
   passphrase_ok = 0;
   w = create_window (want_pass ? 0 : 1);
   gtk_main ();
@@ -454,6 +560,9 @@ gtk_cmd_handler (pinentry_t pe)
   while (gtk_events_pending ())
     gtk_main_iteration ();
 
+  if (confirm_value == CONFIRM_CANCEL || grab_failed)
+    pe->canceled = 1;
+
   pinentry = NULL;
   if (want_pass)
     {
@@ -463,7 +572,7 @@ gtk_cmd_handler (pinentry_t pe)
 	return -1;
     }
   else
-    return confirm_yes;
+    return (confirm_value == CONFIRM_OK) ? 1 : 0;
 }
 
 
@@ -473,8 +582,20 @@ pinentry_cmd_handler_t pinentry_cmd_handler = gtk_cmd_handler;
 int
 main (int argc, char *argv[])
 {
+  static GMemVTable secure_mem =
+    {
+      secentry_malloc,
+      secentry_realloc,
+      secentry_free,
+      NULL,
+      NULL,
+      NULL
+    };
+
+  g_mem_set_vtable (&secure_mem);
+
   pinentry_init (PGMNAME);
-    
+
 #ifdef FALLBACK_CURSES
   if (pinentry_have_display (argc, argv))
     gtk_init (&argc, &argv);
@@ -490,9 +611,9 @@ main (int argc, char *argv[])
       printf(PGMNAME " " VERSION "\n");
       exit(EXIT_SUCCESS);
     }
-  
+
   if (pinentry_loop ())
     return 1;
-  
+
   return 0;
 }
