@@ -33,14 +33,32 @@
 #include <limits.h>
 #include <string.h>
 #include <errno.h>
+#include <time.h>
+#include <sys/types.h>
+#include <sys/stat.h>
+#ifdef HAVE_UTIME_H
+#include <utime.h>
+#endif /*HAVE_UTIME_H*/
 
 #include <memory.h>
 
+#ifdef HAVE_WCHAR_H
+#include <wchar.h>
+#endif /*HAVE_WCHAR_H*/
+
 #include "pinentry.h"
 
+/* FIXME: We should allow configuration of these button labels and in
+   any case use the default_ok, default_cancel values if available.
+   However, I have no clue about curses and localization.  */
 #define STRING_OK "<OK>"
+#define STRING_NOTOK "<No>"
 #define STRING_CANCEL "<Cancel>"
 
+#define USE_COLORS		(has_colors () && COLOR_PAIRS >= 2)
+static short pinentry_color[] = { -1, -1, COLOR_BLACK, COLOR_RED,
+				  COLOR_GREEN, COLOR_YELLOW, COLOR_BLUE,
+				  COLOR_MAGENTA, COLOR_CYAN, COLOR_WHITE };
 static int init_screen;
 
 typedef enum
@@ -48,6 +66,7 @@ typedef enum
     DIALOG_POS_NONE,
     DIALOG_POS_PIN,
     DIALOG_POS_OK,
+    DIALOG_POS_NOTOK,
     DIALOG_POS_CANCEL
   }
 dialog_pos_t;
@@ -72,10 +91,31 @@ struct dialog
   int cancel_y;
   int cancel_x;
   char *cancel;
+  int notok_y;
+  int notok_x;
+  char *notok;
 };
 typedef struct dialog *dialog_t;
 
 
+#ifdef HAVE_NCURSESW
+typedef wchar_t CH;
+#define STRLEN(x) wcslen (x)
+#define ADDCH(x) addnwstr (&x, 1);
+#define CHWIDTH(x) wcwidth (x)
+#define NULLCH L'\0'
+#define NLCH L'\n'
+#define SPCH L' '
+#else
+typedef char CH;
+#define STRLEN(x) strlen (x)
+#define ADDCH(x) addch ((unsigned char) x)
+#define CHWIDTH(x) 1
+#define NULLCH '\0'
+#define NLCH '\n'
+#define SPCH ' '
+#endif
+
 /* Return the next line up to MAXLEN columns wide in START and LEN.
    The first invocation should have 0 as *LEN.  If the line ends with
    a \n, it is a normal line that will be continued.  If it is a '\0'
@@ -83,40 +123,95 @@ typedef struct dialog *dialog_t;
    there is a forced line break.  A full line is returned and will be
    continued in the next line.  */
 static void
-collect_line (int maxlen, char **start_p, int *len_p)
+collect_line (int maxwidth, wchar_t **start_p, int *len_p)
 {
   int last_space = 0;
   int len = *len_p;
-  char *end;
+  int width = 0;
+  CH *end;
 
   /* Skip to next line.  */
   *start_p += len;
   /* Skip leading space.  */
-  while (**start_p == ' ')
+  while (**start_p == SPCH)
     (*start_p)++;
 
   end = *start_p;
   len = 0;
 
-  while (len < maxlen - 1 && *end && *end != '\n')
+  while (width < maxwidth - 1 && *end != NULLCH && *end != NLCH)
     {
       len++;
       end++;
-      if (*end == ' ')
+      if (*end == SPCH)
 	last_space = len;
+      width += CHWIDTH (*end);
     }
 
-  if (*end && *end != '\n' && last_space != 0)
+  if (*end != NULLCH && *end != NLCH && last_space != 0)
     {
       /* We reached the end of the available space, but still have
 	 characters to go in this line.  We can break the line into
 	 two parts at a space.  */
       len = last_space;
-      (*start_p)[len] = '\n';
+      (*start_p)[len] = NLCH;
     }
   *len_p = len + 1;
 }
 
+#ifdef HAVE_NCURSESW
+static CH *
+utf8_to_local (char *lc_ctype, char *string)
+{
+  mbstate_t ps;
+  size_t len;
+  char *local;
+  const char *p;
+  wchar_t *wcs = NULL;
+  char *old_ctype = NULL;
+
+  local = pinentry_utf8_to_local (lc_ctype, string);
+  if (!local)
+    return NULL;
+
+  old_ctype = strdup (setlocale (LC_CTYPE, NULL));
+  setlocale (LC_CTYPE, lc_ctype? lc_ctype : "");
+
+  p = local;
+  memset (&ps, 0, sizeof(mbstate_t));
+  len = mbsrtowcs (NULL, &p, strlen (string), &ps);
+  if (len == (size_t)-1)
+    {
+      free (local);
+      goto leave;
+    }
+  wcs = calloc (len + 1, sizeof(wchar_t));
+  if (!wcs)
+    {
+      free (local);
+      goto leave;
+    }
+
+  p = local;
+  memset (&ps, 0, sizeof(mbstate_t));
+  mbsrtowcs (wcs, &p, len, &ps);
+
+ leave:
+  if (old_ctype)
+    {
+      setlocale (LC_CTYPE, old_ctype);
+      free (old_ctype);
+    }
+
+  return wcs;
+}
+#else
+static CH *
+utf8_to_local (const char *lc_ctype, const char *string)
+{
+  return pinentry_utf8_to_local (lc_ctype, string);
+}
+#endif
 
 static int
 dialog_create (pinentry_t pinentry, dialog_t dialog)
@@ -130,16 +225,15 @@ dialog_create (pinentry_t pinentry, dialog_t dialog)
   int xpos;
   int description_x = 0;
   int error_x = 0;
-  char *description = NULL;
-  char *error = NULL;
-  char *prompt = NULL;
+  CH *description = NULL;
+  CH *error = NULL;
+  CH *prompt = NULL;
 
 #define COPY_OUT(what)							\
   do									\
     if (pinentry->what)							\
       {									\
-        what = pinentry_utf8_to_local (pinentry->lc_ctype,		\
-				       pinentry->what);			\
+        what = utf8_to_local (pinentry->lc_ctype, pinentry->what);	\
         if (!what)							\
 	  {								\
 	    err = 1;							\
@@ -181,7 +275,14 @@ dialog_create (pinentry_t pinentry, dialog_t dialog)
   while (0)
 
   MAKE_BUTTON (ok, STRING_OK);
-  MAKE_BUTTON (cancel, STRING_CANCEL);
+  if (!pinentry->one_button)
+    MAKE_BUTTON (cancel, STRING_CANCEL);
+  else
+    dialog->cancel = NULL;
+  if (!pinentry->one_button && pinentry->notok)
+    MAKE_BUTTON (notok, STRING_NOTOK);
+  else
+    dialog->notok = NULL;
 
   getmaxyx (stdscr, size_y, size_x);
 
@@ -189,7 +290,7 @@ dialog_create (pinentry_t pinentry, dialog_t dialog)
   y = 1;		/* Top frame.  */
   if (description)
     {
-      char *start = description;
+      CH *start = description;
       int len = 0;
 
       do
@@ -207,7 +308,7 @@ dialog_create (pinentry_t pinentry, dialog_t dialog)
     {
       if (error)
 	{
-	  char *p = error;
+	  CH *p = error;
 	  int err_x = 0;
 
 	  while (*p)
@@ -262,18 +363,24 @@ dialog_create (pinentry_t pinentry, dialog_t dialog)
 
       new_x = MIN_PINENTRY_LENGTH;
       if (prompt)
-	new_x += strlen (prompt) + 1;	/* One space after prompt.  */
+	{
+	  new_x += STRLEN (prompt) + 1;	/* One space after prompt.  */
+	}
       if (new_x > size_x - 4)
 	new_x = size_x - 4;
       if (new_x > x)
 	x = new_x;
     }
-  /* We position the buttons after the first and second third of the
-     width.  Account for rounding.  */
-  if (x < 2 * strlen (dialog->ok))
-    x = 2 * strlen (dialog->ok);
-  if (x < 2 * strlen (dialog->cancel))
-    x = 2 * strlen (dialog->cancel);
+  /* We position the buttons after the first, second and third fourth
+     of the width.  Account for rounding.  */
+  if (x < 3 * strlen (dialog->ok))
+    x = 3 * strlen (dialog->ok);
+  if (dialog->cancel)
+    if (x < 3 * strlen (dialog->cancel))
+      x = 3 * strlen (dialog->cancel);
+  if (dialog->notok)
+    if (x < 3 * strlen (dialog->notok))
+      x = 3 * strlen (dialog->notok);
 
   /* Add the frame.  */
   x += 4;
@@ -306,7 +413,7 @@ dialog_create (pinentry_t pinentry, dialog_t dialog)
   ypos++;
   if (description)
     {
-      char *start = description;
+      CH *start = description;
       int len = 0;
 
       do
@@ -318,9 +425,11 @@ dialog_create (pinentry_t pinentry, dialog_t dialog)
 	  addch (' ');
 	  collect_line (size_x - 4, &start, &len);
 	  for (i = 0; i < len - 1; i++)
-	    addch ((unsigned char) start[i]);
-	  if (start[len - 1] && start[len - 1] != '\n')
-	    addch ((unsigned char) start[len - 1]);
+	    {
+	      ADDCH (start[i]);
+	    }
+	  if (start[len - 1] != NULLCH && start[len - 1] != NLCH)
+	    ADDCH (start[len - 1]);
 	  ypos++;
 	}
       while (start[len - 1]);
@@ -334,7 +443,7 @@ dialog_create (pinentry_t pinentry, dialog_t dialog)
 
       if (error)
 	{
-	  char *p = error;
+	  CH *p = error;
 	  i = 0;
 
 	  while (*p)
@@ -342,18 +451,24 @@ dialog_create (pinentry_t pinentry, dialog_t dialog)
 	      move (ypos, xpos);
 	      addch (ACS_VLINE);
 	      addch (' ');
-	      if (has_colors () && COLOR_PAIRS >= 1)
-		attron (COLOR_PAIR(1) | A_BOLD);
+	      if (USE_COLORS && pinentry->color_so != PINENTRY_COLOR_NONE)
+		{
+		  attroff (COLOR_PAIR (1) | (pinentry->color_fg_bright ? A_BOLD : 0));
+		  attron (COLOR_PAIR (2) | (pinentry->color_so_bright ? A_BOLD : 0));
+		}
 	      else
 		standout ();
-	      for (;*p && *p != '\n'; p++)
+	      for (;*p && *p != NLCH; p++)
 		if (i < x - 4)
 		  {
 		    i++;
-		    addch ((unsigned char) *p);
+		    ADDCH (*p);
 		  }
-	      if (has_colors () && COLOR_PAIRS >= 1)
-		attroff (COLOR_PAIR(1) | A_BOLD);
+	      if (USE_COLORS && pinentry->color_so != PINENTRY_COLOR_NONE)
+		{
+		  attroff (COLOR_PAIR (2) | (pinentry->color_so_bright ? A_BOLD : 0));
+		  attron (COLOR_PAIR (1) | (pinentry->color_fg_bright ? A_BOLD : 0));
+		}
 	      else
 		standend ();
 	      if (*p == '\n')
@@ -375,14 +490,16 @@ dialog_create (pinentry_t pinentry, dialog_t dialog)
       dialog->pin_size = x - 4;
       if (prompt)
 	{
-	  char *p = prompt;
-	  i = strlen (prompt);
+	  CH *p = prompt;
+	  i = STRLEN (prompt);
 	  if (i > x - 4 - MIN_PINENTRY_LENGTH)
 	    i = x - 4 - MIN_PINENTRY_LENGTH;
 	  dialog->pin_x += i + 1;
 	  dialog->pin_size -= i + 1;
 	  while (i-- > 0)
-	    addch ((unsigned char) *(p++));
+	    {
+	      ADDCH (*(p++));
+	    }
 	  addch (' ');
 	}
       for (i = 0; i < dialog->pin_size; i++)
@@ -394,16 +511,40 @@ dialog_create (pinentry_t pinentry, dialog_t dialog)
     }
   move (ypos, xpos);
   addch (ACS_VLINE);
-  dialog->ok_y = ypos;
-  /* Calculating the left edge of the left button, rounding down.  */
-  dialog->ok_x = xpos + 2 + ((x - 4) / 2 - strlen (dialog->ok)) / 2;
-  move (dialog->ok_y, dialog->ok_x);
-  addstr (dialog->ok);
-  dialog->cancel_y = ypos;
-  /* Calculating the left edge of the right button, rounding up.  */
-  dialog->cancel_x = xpos + x - 2 - ((x - 4) / 2 + strlen (dialog->cancel)) / 2;
-  move (dialog->cancel_y, dialog->cancel_x);
-  addstr (dialog->cancel);
+
+  if (dialog->cancel || dialog->notok)
+    {
+      dialog->ok_y = ypos;
+      /* Calculating the left edge of the left button, rounding down.  */
+      dialog->ok_x = xpos + 2 + ((x - 4) / 3 - strlen (dialog->ok)) / 2;
+      move (dialog->ok_y, dialog->ok_x);
+      addstr (dialog->ok);
+
+      if (dialog->notok)
+	{
+	  dialog->notok_y = ypos;
+	  /* Calculating the left edge of the middle button, rounding up.  */
+	  dialog->notok_x = xpos + x / 2 - strlen (dialog->notok) / 2;
+	  move (dialog->notok_y, dialog->notok_x);
+	  addstr (dialog->notok);
+	}
+      if (dialog->cancel)
+	{
+	  dialog->cancel_y = ypos;
+	  /* Calculating the left edge of the right button, rounding up.  */
+	  dialog->cancel_x = xpos + x - 2 - ((x - 4) / 3 + strlen (dialog->cancel)) / 2;
+	  move (dialog->cancel_y, dialog->cancel_x);
+	  addstr (dialog->cancel);
+	}
+    }
+  else
+    {
+      dialog->ok_y = ypos;
+      /* Calculating the left edge of the OK button, rounding down.  */
+      dialog->ok_x = xpos + x / 2 - strlen (dialog->ok) / 2;
+      move (dialog->ok_y, dialog->ok_x);
+      addstr (dialog->ok);
+    }
 
  out:
   if (description)
@@ -445,9 +586,19 @@ dialog_switch_pos (dialog_t diag, dialog_pos_t new_pos)
 	  move (diag->ok_y, diag->ok_x);
 	  addstr (diag->ok);
 	  break;
+	case DIALOG_POS_NOTOK:
+          if (diag->notok)
+            {
+              move (diag->notok_y, diag->notok_x);
+              addstr (diag->notok);
+            }
+	  break;
 	case DIALOG_POS_CANCEL:
-	  move (diag->cancel_y, diag->cancel_x);
-	  addstr (diag->cancel);
+          if (diag->cancel)
+            {
+              move (diag->cancel_y, diag->cancel_x);
+              addstr (diag->cancel);
+            }
 	  break;
 	default:
 	  break;
@@ -460,18 +611,34 @@ dialog_switch_pos (dialog_t diag, dialog_pos_t new_pos)
 	  set_cursor_state (1);
 	  break;
 	case DIALOG_POS_OK:
+	  set_cursor_state (0);
 	  move (diag->ok_y, diag->ok_x);
 	  standout ();
 	  addstr (diag->ok);
 	  standend ();
-	  set_cursor_state (0);
+	  move (diag->ok_y, diag->ok_x);
+	  break;
+	case DIALOG_POS_NOTOK:
+          if (diag->notok)
+            {
+              set_cursor_state (0);
+              move (diag->notok_y, diag->notok_x);
+              standout ();
+              addstr (diag->notok);
+              standend ();
+              move (diag->notok_y, diag->notok_x);
+            }
 	  break;
 	case DIALOG_POS_CANCEL:
-	  move (diag->cancel_y, diag->cancel_x);
-	  standout ();
-	  addstr (diag->cancel);
-	  standend ();
-	  set_cursor_state (0);
+          if (diag->cancel)
+            {
+              set_cursor_state (0);
+              move (diag->cancel_y, diag->cancel_x);
+              standout ();
+              addstr (diag->cancel);
+              standend ();
+              move (diag->cancel_y, diag->cancel_x);
+            }
 	  break;
 	case DIALOG_POS_NONE:
 	  set_cursor_state (0);
@@ -546,6 +713,17 @@ dialog_run (pinentry_t pinentry, const char *tty_name, const char *tty_type)
   SCREEN *screen = 0;
   int done = 0;
   char *pin_utf8;
+#ifdef HAVE_NCURSESW
+  char *old_ctype = NULL;
+
+  if (pinentry->lc_ctype)
+    {
+      old_ctype = strdup (setlocale (LC_CTYPE, NULL));
+      setlocale (LC_CTYPE, pinentry->lc_ctype);
+    }
+  else
+    setlocale (LC_CTYPE, "");
+#endif
 
   /* Open the desired terminal if necessary.  */
   if (tty_name)
@@ -579,15 +757,29 @@ dialog_run (pinentry_t pinentry, const char *tty_name, const char *tty_type)
   nonl ();		/* Tell curses not to do NL->CR/NL on output.  */
   cbreak ();		/* Take input chars one at a time, no wait for \n.  */
   noecho ();		/* Don't echo input - in color.  */
-  refresh ();
 
   if (has_colors ())
     {
       start_color ();
+      use_default_colors ();
 
-      if (COLOR_PAIRS >= 1)
-	init_pair (1, COLOR_RED, COLOR_BLACK);
+      if (pinentry->color_so == PINENTRY_COLOR_DEFAULT)
+	{
+	  pinentry->color_so = PINENTRY_COLOR_RED;
+	  pinentry->color_so_bright = 1;
+	}
+      if (COLOR_PAIRS >= 2)
+	{
+	  init_pair (1, pinentry_color[pinentry->color_fg],
+		     pinentry_color[pinentry->color_bg]);
+	  init_pair (2, pinentry_color[pinentry->color_so],
+		     pinentry_color[pinentry->color_bg]);
+
+	  bkgd (COLOR_PAIR (1));
+	  attron (COLOR_PAIR (1) | (pinentry->color_fg_bright ? A_BOLD : 0));
+	}
     }
+  refresh ();
 
   /* XXX */
   if (dialog_create (pinentry, &diag))
@@ -610,9 +802,15 @@ dialog_run (pinentry_t pinentry, const char *tty_name, const char *tty_type)
 	      if (diag.pin)
 		dialog_switch_pos (&diag, DIALOG_POS_PIN);
 	      break;
-	    case DIALOG_POS_CANCEL:
+	    case DIALOG_POS_NOTOK:
 	      dialog_switch_pos (&diag, DIALOG_POS_OK);
 	      break;
+	    case DIALOG_POS_CANCEL:
+	      if (diag.notok)
+		dialog_switch_pos (&diag, DIALOG_POS_NOTOK);
+	      else
+		dialog_switch_pos (&diag, DIALOG_POS_OK);
+	      break;
 	    default:
 	      break;
 	    }
@@ -626,6 +824,12 @@ dialog_run (pinentry_t pinentry, const char *tty_name, const char *tty_type)
 	      dialog_switch_pos (&diag, DIALOG_POS_OK);
 	      break;
 	    case DIALOG_POS_OK:
+	      if (diag.notok)
+		dialog_switch_pos (&diag, DIALOG_POS_NOTOK);
+	      else
+		dialog_switch_pos (&diag, DIALOG_POS_CANCEL);
+	      break;
+	    case DIALOG_POS_NOTOK:
 	      dialog_switch_pos (&diag, DIALOG_POS_CANCEL);
 	      break;
 	    default:
@@ -640,6 +844,12 @@ dialog_run (pinentry_t pinentry, const char *tty_name, const char *tty_type)
 	      dialog_switch_pos (&diag, DIALOG_POS_OK);
 	      break;
 	    case DIALOG_POS_OK:
+	      if (diag.notok)
+		dialog_switch_pos (&diag, DIALOG_POS_NOTOK);
+	      else
+		dialog_switch_pos (&diag, DIALOG_POS_CANCEL);	      
+	      break;
+	    case DIALOG_POS_NOTOK:
 	      dialog_switch_pos (&diag, DIALOG_POS_CANCEL);
 	      break;
 	    case DIALOG_POS_CANCEL:
@@ -653,7 +863,7 @@ dialog_run (pinentry_t pinentry, const char *tty_name, const char *tty_type)
 	    }
 	  break;
   
-	case '\e':
+	case '\005':
 	  done = -2;
 	  break;
 
@@ -664,6 +874,9 @@ dialog_run (pinentry_t pinentry, const char *tty_name, const char *tty_type)
 	    case DIALOG_POS_OK:
 	      done = 1;
 	      break;
+	    case DIALOG_POS_NOTOK:
+	      done = -1;
+	      break;
 	    case DIALOG_POS_CANCEL:
 	      done = -2;
 	      break;
@@ -684,30 +897,76 @@ dialog_run (pinentry_t pinentry, const char *tty_name, const char *tty_type)
   if (screen)
     delscreen (screen);
 
+#ifdef HAVE_NCURSESW
+  if (old_ctype)
+    {
+      setlocale (LC_CTYPE, old_ctype);
+      free (old_ctype);
+    }
+#endif
   if (ttyfi)
     fclose (ttyfi);
   if (ttyfo)
     fclose (ttyfo);
   /* XXX Factor out into dialog_release or something.  */
   free (diag.ok);
-  free (diag.cancel);
+  if (diag.cancel)
+    free (diag.cancel);
+  if (diag.notok)
+    free (diag.notok);
 
-  pinentry->locale_err = 1;
-  pin_utf8 = pinentry_local_to_utf8 (pinentry->lc_ctype, pinentry->pin, 1);
-  if (pin_utf8)
+  if (pinentry->pin)
     {
-      pinentry_setbufferlen (pinentry, strlen (pin_utf8) + 1);
-      if (pinentry->pin)
-	strcpy (pinentry->pin, pin_utf8);
-      secmem_free (pin_utf8);
-      pinentry->locale_err = 0;
+      pinentry->locale_err = 1;
+      pin_utf8 = pinentry_local_to_utf8 (pinentry->lc_ctype, pinentry->pin, 1);
+      if (pin_utf8)
+	{
+	  pinentry_setbufferlen (pinentry, strlen (pin_utf8) + 1);
+	  if (pinentry->pin)
+	    strcpy (pinentry->pin, pin_utf8);
+	  secmem_free (pin_utf8);
+	  pinentry->locale_err = 0;
+	}
     }
 
+  if (done == -2)
+    pinentry->canceled = 1;
+
   return diag.pin ? (done < 0 ? -1 : diag.pin_len) : (done < 0 ? 0 : 1);
 }
 
+
+/* If a touch has been registered, touch that file.  */
+static void
+do_touch_file (pinentry_t pinentry)
+{
+#ifdef HAVE_UTIME_H
+  struct stat st;
+  time_t tim;
+
+  if (!pinentry->touch_file || !*pinentry->touch_file)
+    return;
+
+  if (stat (pinentry->touch_file, &st))
+    return; /* Oops.  */
+
+  /* Make sure that we actually update the mtime.  */
+  while ( (tim = time (NULL)) == st.st_mtime )
+    sleep (1);
+
+  /* Update but ignore errors as we can't do anything in that case.
+     Printing error messages may even clubber the display further. */
+  utime (pinentry->touch_file, NULL);
+#endif /*HAVE_UTIME_H*/
+}
+
+
 int
 curses_cmd_handler (pinentry_t pinentry)
 {
-  return dialog_run (pinentry, pinentry->ttyname, pinentry->ttytype);
+  int rc;
+
+  rc = dialog_run (pinentry, pinentry->ttyname, pinentry->ttytype);
+  do_touch_file (pinentry);
+  return rc;
 }
